#include <json.hpp>
#include <lexer.hpp>
#include <sstream>
#include <format>

namespace json {


    struct serializable {
        virtual json::value serialize() const = 0;
    };

    struct deserializable {
        virtual void deserialize(const json::value&) = 0;
    };





    std::string format_error(std::string base, std::string source, int index) {
        std::ostringstream s;
        int counter = 0;
        int line = 1;
        int column = 0;
        std::string lastline = "";
        std::string whitespace = "";

        for (auto c: source) {
            if (counter == index) {
                break;
            }

            if (c == '\n') {
                line++;
                column = 0;
                lastline = "";
                whitespace = "";
            } else if (c == '\t') {
                column++;
                lastline += "   ";
                whitespace += "   ";
            } else {
                column++;
                lastline += c;
                whitespace += " ";
            }
            counter++;
        }
        while (counter < source.size()) {
            auto c = source[counter];
            if (c == '\n') {
                break;
            }
            lastline += c;
            counter++;
        }

        // TODO: Migrate the below code bits to std::format

        s << base << " at line " << line << ", column " << column << std::endl;
        s << lastline << std::endl;
        s << whitespace << "^";

        return s.str();
    }



    json::value& value::operator=(double in)
    {
        type = value_type::number;
        number = in;
        return *this;
    }

    json::value& value::operator=(const std::string& in)
    {
        type = value_type::string;
        string = in;
        return *this;
    }

    json::value& value::operator=(bool in)
    {
        type = value_type::boolean;
        boolean = in;
        return *this;
    }

    value& value::operator=(const std::vector<value>& in)
    {
        type = value_type::array;
        array = in;
        return *this;
    }

    value& value::operator=(const std::map<std::string, value>& in)
    {
        type = value_type::object;
        object = in;
        return *this;
    }

    value::operator double() const { return number.value(); }

    value::operator std::string() const { return string.value(); }

    value::operator bool() const { return boolean.value(); }

    value::operator std::vector<value>() const { return this->value::array.value(); }

    value::operator std::map<std::string, value>() const { return object.value(); }

    struct json::object json::value::as_object_value() const {
        return json::object(object.value());
    }

    struct json::array json::value::as_array_value() const {
        return json::array(array.value());
    }

    struct json::string json::value::as_string_value() const {
        return json::string(string.value());
    }

    struct json::number json::value::as_number_value() const {
        return json::number(number.value());
    }

    struct json::boolean json::value::as_boolean_value() const {
        return json::boolean(boolean.value());
    }

    std::map<std::string, value> value::as_object() const
    {
        return object.value();
    }

    std::vector<value> value::as_array() const
    { return array.value(); }

    float value::as_float() const
    { return number.value(); }

    float value::as_float_or(float def) const
    { return number.value_or(def); }

    double value::as_double() const { return number.value();}

    double value::as_double_or(double def) const { return number.value_or(def); }

    std::string value::as_string() const
    { return string.value();}

    std::string value::as_string_or(const std::string& def) const
    { return string.value_or(def);}



    void value::convert_descendants() {
        if (type == value_type::object)
        {
            for (auto& [k, v] : object.value())
            {
                if (v.type == value_type::object)
                {
                    auto converted = v.as_object_value();
                    converted.convert_descendants();
                    object.value().emplace(k, converted);
                }

                if (v.type == value_type::array)
                {
                    auto converted = v.as_array_value();
                    converted.convert_descendants();
                    object.value().emplace(k, converted);
                }


                if (v.type == value_type::string)
                    object.value().emplace(k, v.as_string_value());

                if (v.type == value_type::number)
                    object.value().emplace(k, v.as_number_value());

                if (v.type == value_type::boolean)
                    object.value().emplace(k, v.as_boolean_value());
            }
        }

        if (type == value_type::array)
        {
            for (int i = 0; i < array.value().size(); ++i)
            {
                auto v = array.value()[i];
                if (v.type == value_type::object) {
                    auto converted = v.as_object_value();
                    converted.convert_descendants();
                    array.value()[i] = converted;
                } else if (v.type == value_type::array) {
                    auto converted = v.as_array_value();
                    converted.convert_descendants();
                    array.value()[i] = converted;
                } else if (v.type == value_type::string)
                    array.value()[i] = v.as_string_value();
                else if (v.type == value_type::number)
                    array.value()[i] = v.as_number_value();
                else if (v.type == value_type::boolean)
                    array.value()[i] = v.as_boolean_value();
            }
        }
    }

    string::operator std::string()
    {
        return this->value::string.value();
    }


    value& object::operator[](const std::string& key)
    {
        return this->value::object.value()[key];
    }

    const value & object::operator[](const std::string &key) const { return value::object->at(key);}


    value& array::operator[](int key)
    {
        return this->value::array.value()[key];
    }

    json::value_type array::type_of(int index) {
        return this->value::array->at(index).type;
    }

    value &array::operator+=(const value &v) {
        push_back(v);
        return *this;
    }

    array::const_iterator array::begin() const { return value::array->cbegin(); }

    array::const_iterator array::end() const { return value::array->cend(); }

    array::const_iterator array::cbegin() const { return value::array->cbegin();}

    array::const_iterator array::cend() const { return value::array->cend();}

    void array::push_back(const value &element) {
        value::array->push_back(element);
    }

    std::tuple<std::vector<json::token>, std::string> lex(std::string raw_json) {
        using namespace lexers;
        std::vector<json::token> tokens;
        // All tokens will embed a pointer to the raw JSON for debugging purposes
        auto original_copy = std::make_shared<std::string>(raw_json);

        auto generic_lexers = {lex_syntax, lex_string, lex_number, lex_null, lex_true, lex_false};

        for (int i = 0; i < raw_json.length(); i++) {
            // Skip past whitespace
            if (auto new_index = lex_whitespace(raw_json, i); i != new_index) {
                i = new_index - 1;
                continue;
            }

            auto found = false;
            for (auto lexer: generic_lexers) {
                if (auto [token, new_index, error] = lexer(raw_json, i); i != new_index) {
                    // Error while lexing, return early
                    if (error.length()) {
                        return std::make_tuple(std::vector<json::token>{}, error);
                    }

                    // Store reference to the original source
                    token.full_source = original_copy;
                    tokens.push_back(token);
                    i = new_index - 1;
                    found = true;
                    break;
                }
            }

            if (found) {
                continue;
            }
            return std::make_tuple(std::vector<json::token>{}, format_error("Unable to lex", raw_json, i));
        }
        return {tokens, ""};
    }

    // It's very annoying when languages doesn't give you
    // stringifier methods for enums by default for debugging.
    // There are ways to do this with reflection but it seems hairy.
    // There's a better procedure IIRC.
    std::string token_type_tostring(token_type tok)
    {
        switch(tok) {
            case token_type::string: return "String";
            case token_type::number: return "Number";
            case token_type::syntax: return "Syntax";
            case token_type::boolean: return "Boolean";
            case token_type::null: return "Null";
        }
    }

    std::string format_parse_error(std::string base, json::token token)
    {
        std::ostringstream s;
        s << "Unexpected token '" << token.value << "', type' '"
          << token_type_tostring(token.type)     << "', index'";
        s << std::endl << base;
        return format_error(s.str(), *token.full_source, token.location);
    }

    std::tuple<std::vector<json::value>, int, std::string> parse_array(std::vector<json::token> tokens, int index) {
        std::vector<json::value> children = {};
        while (index < tokens.size()) {
            auto t = tokens[index];
            if (t.type == token_type::syntax) {
                if (t.value == "]") {
                    return {children, index + 1, ""};
                }
                if (t.value == ",") {
                    index++;
                    t = tokens[index];
                } else if (children.size() > 0) {
                    return {{}, index,
                            format_parse_error("Expected comma after element in array", t)};
                }
            }

            auto [child, new_index, error] = parse(tokens, index);
            if (error.size()) { return {{}, index, error}; }
            children.push_back(child);
            index = new_index;
        }
        return {
                {}, index,
                format_parse_error("Unexpected EOF while parsing array", tokens[index])};

    }

    std::tuple<std::map<std::string, value>, int, std::string> parse_object(std::vector<token> tokens, int index) {
        std::map<std::string, value> values = {};
        while (index < tokens.size()) {
            auto t = tokens[index];
            if (t.type == token_type::syntax) {
                if (t.value == "}") {
                    return {values, index + 1, ""};
                }
                if (t.value == ",") {
                    index++;
                    t = tokens[index];
                } else if (values.size() > 0) {
                    return {{}, index,
                            format_parse_error("Expected comma after element in object", t)
                    };
                } else {
                    return {{}, index,
                            format_parse_error("Expected key-value pair or closing brace in object", t)
                    };
                }
            }
            auto [key, new_index, error] = parse(tokens, index);
            if (!error.empty())
            {
                return {{}, index, error};
            }

            if (key.type != value_type::string) {
                return {{}, index,
                        format_parse_error("Expected string key in object", t)};
            }

            index = new_index;
            t = tokens[index];

            if (!(t.type == token_type::syntax && t.value == ":")) {
                return {{}, index,
                        format_parse_error("Expected colon after key in object", t)};
            }
            index++;
            t = tokens[index];

            auto [value, new_index1, error1] = parse(tokens, index);
            if (error1.size()) {
                return {{}, index, error1};
            }

            values[key.string.value()] = value;
            index = new_index1;
        }
        return {values, index+1, ""};
    }

    std::tuple<json::value, int, std::string> parse(std::vector<json::token> tokens, int index) {
        auto token = tokens[index];
        switch(token.type) {
            case token_type::number: {
                double n = std::stod(token.value);
                return {json::number(n), index + 1, ""};
            }
            case token_type::boolean:
                return {json::boolean(token.value == "true"), index + 1, ""};
            case token_type::null:
                return {json::value(), index+1, ""};
            case token_type::string:
                return {json::value(token.value), index+1, ""};
            case token_type::syntax: {
                if (token.value == "[") {
                    auto [array, new_index, error] = parse_array(tokens, index + 1);
                    return {json::value(array), new_index, error};
                }
                if (token.value == "{") {
                    auto [object, new_index, error] = parse_object(tokens, index + 1);
                    return {json::value(object), new_index, error};
                }
            }
            default:
                return {json::value(), index, format_parse_error("Failed to parse", token)};
        }
    }

    std::tuple<json::value, std::string> parse(std::string source) {
        auto [tokens, error] = json::lex(source);
        if (error.size())
        {
            return std::make_tuple(json::value{}, error);
        }

        auto [ast, _, error1] = json::parse(tokens);
        return {ast, error1};
    }

    std::string deparse(json::value v, std::string whitespace) {
        switch(v.type) {
            case json::value_type::string:
                return "\"" + v.string.value() + "\"";
            case json::value_type::boolean:
                return (v.boolean.value() ? "true" : "false");
            case json::value_type::number:
                return std::to_string(v.number.value());
            case json::value_type::null:
                return "null";
            case json::value_type::array: {
                std::string s = "[\n";
                auto a = v.array.value();
                for (int i = 0; i < a.size(); i++) {
                    auto value = a[i];
                    s += whitespace + " " + deparse(value, whitespace + " ");
                    if (i < a.size() - 1) {
                        s += ",";
                    }
                    s += "\n";
                }
                return s + whitespace + "]";
            }
            case json::value_type::object: {
                std::string s = "{\n";
                auto values = v.object.value();
                auto i = 0;
                for (auto const &[key, value] : values) {
                    s += whitespace + " " + "\"" + key + "\":" + deparse(value, whitespace + " ");

                    if (i < values.size() - 1) {
                        s += ",";
                    }
                    s += "\n";
                    i++;
                }
                return s + whitespace + "}";
            }
        }
    }

    value::value(): type(value_type::null) { }

    value::value(double v): type(value_type::number), number(v) {}

    value::value(const std::string &v): string(v), type(value_type::string) {}

    value::value(const std::vector<value> &v): array(v), type(value_type::array) {}

    value::value(const std::map<std::string, value> &v): object(v), type(value_type::object) {}

    value::value(bool v) : boolean(v), type(value_type::boolean) {}

    value::value(int v) : number(v), type(value_type::number) {}

    value::value(float v) : number(v), type(value_type::number) {}

    value &value::at(const std::string &key) {
        assert(type == value_type::object && "Not a JSON object");
        return this->value::object->at(key);
    }

    value &value::operator[](const std::string &key) {
        assert(type == value_type::object);
        // TODO: utilize std::map operator
        return value::object->at(key);
    }

    const value &value::operator[](const std::string &key) const {
        assert(type == value_type::object);
        // TODO: utilize std::map operator
        return value::object->at(key);
    }

    bool value::contains(const std::string &key) {
        assert(type == value_type::object);
        return this->value::object->contains(key);
    }

    bool value::is(value_type type) const { return type == this->type; }

    bool value::is_number() const { return is(value_type::number);}

    bool value::is_string() const { return is(value_type::string);}

    bool value::is_array() const { return is(value_type::array); }

    bool value::is_object() const { return is(value_type::object); }

    bool value::is_boolean() const { return is(value_type::boolean); }

    bool value::is_null() const { return is(value_type::null);}

    value &value::operator[](int index) {
        assert(is_array());
        return value::array->operator[](index);
    }

    value &value::at(int index) {
        assert(type == value_type::array && "Not a JSON array!");

        return value::array->at(index);
    }

    const value& value::operator[](int index) const
    {
        assert(type == value_type::array && "Not a JSON array!");
        return value::array->operator[](index);
    }


    struct string string(const std::string &text) {
        struct string out;
        out.type = value_type::string;
        out.value::string = text;
        return out;
    }

    struct number number(double input) {
        struct number out;
        out.type = value_type::number;
        out.value::number = input;
        return out;
    }

    struct boolean boolean(bool input) {
        struct boolean out;
        out.type = value_type::boolean;
        out.number = input;
        return out;
    }

    array::array(const std::vector<value>& v)
    {

        this->type = value_type::array;
        this->value::array = std::vector<value>();

        for (auto& elem: v)
            this->value::array->push_back(elem);

    }

    /*struct array array()
    {
        struct json::array out;
        out.type = value_type::array;
        out.array = std::vector<value>();
        return out;
    }

    struct object object(std::map<std::string, value> input)
    {
        struct object out;
        out.type = value_type::object;
        out.value::object = input;
        return out;
    }

    struct object object()
    {
        struct object out;
        out.type = value_type::object;
        out.value::object = std::map<std::string, value>();
        return out;
    }*/

    bool object::contains(const std::string &key) {
        return this->value::object->contains(key);
    }

    void object::insert(const std::pair<std::string, value> &kvp) {
        value::object->insert(kvp);
    }

    void object::insert(const std::string &key, const value &val) {
        value::object->insert({key, val});
    }

    json::object & object::operator+=(const std::pair<std::string, value> &kvp) {
        insert(kvp);
        return *this;
    }

    value &object::at(const std::string &key) {
        return this->value::object.value()[key];
    }

    json::value_type object::type_of(const std::string &key) {
        return this->value::object->at(key).type;
    }

    object::object() : value(std::map<std::string, value>{}) {}

    object::object(const std::map<std::string, value> &v) : value(v) {}

    array::array(): value(std::vector<value>{}) {}

    size_t array::size() const { return value::array->size();}

    number::number() : value(0.0) {}

    number::number(double v) : value(v) {}

    boolean::boolean() : value(false) {}

    boolean::boolean(bool v) : value(v) {}
}
